# Tamito KAJIYAMA <19 August 2009>

# changes for DXF ouput Thomas Lachmann 2013
# - DXF output instead of SVG
# - color mapping to ACI colors
# - rendersize calculated
# - frustum culling added


from freestyle import *
from logical_operators import *
from shaders import *
import bpy
import os
import dxfLibrary130
import freestyle
import math
from mathutils import *

CLIP_LEFT = 1
CLIP_RIGHT = 2
CLIP_TOP = 3
CLIP_BOTTOM = 4

ref_X =  95.047
ref_Y = 100.000
ref_Z = 108.883
def _rgb_to_xyz(R, G, B):
    var_R = R
    var_G = G
    var_B = B
    if var_R > 0.04045:
        var_R = ( ( var_R + 0.055 ) / 1.055 ) ** 2.4
    else:
        var_R /= 12.92

    if var_G > 0.04045:
        var_G = ( ( var_G + 0.055 ) / 1.055 ) ** 2.4
    else:
        var_G /= 12.92
    if var_B > 0.04045:
        var_B = ( ( var_B + 0.055 ) / 1.055 ) ** 2.4
    else:
        var_B /= 12.92

    var_R *= 100
    var_G *= 100
    var_B *= 100

    #Observer. = 2 deg, Illuminant = D65
    X = var_R * 0.4124 + var_G * 0.3576 + var_B * 0.1805
    Y = var_R * 0.2126 + var_G * 0.7152 + var_B * 0.0722
    Z = var_R * 0.0193 + var_G * 0.1192 + var_B * 0.9505

    return X,Y,Z

def _xyz_to_cielab(X, Y, Z):
    var_X = X / ref_X
    var_Y = Y / ref_Y
    var_Z = Z / ref_Z

    if var_X > 0.008856:
        var_X **= ( 1./3. )
    else:
        var_X = ( 7.787 * var_X ) + ( 16. / 116. )
    if var_Y > 0.008856:
        var_Y **= ( 1./3. )
    else:
        var_Y = ( 7.787 * var_Y ) + ( 16. / 116. )
    if var_Z > 0.008856:
        var_Z **= ( 1./3. )
    else:
        var_Z = ( 7.787 * var_Z ) + ( 16. / 116. )

    CIE_L = ( 116 * var_Y ) - 16.
    CIE_a = 500. * ( var_X - var_Y )
    CIE_b = 200. * ( var_Y - var_Z )

    return CIE_L, CIE_a, CIE_b

def _rgb_to_cielab(R, G, B):
    X,Y,Z = _rgb_to_xyz(R,G,B)
    return _xyz_to_cielab(X,Y,Z)

	
def col2DXF(r,g,b):
    from dxfColorMap import color_map
    dxfColors = [(tuple(color)) for idx, color in color_map.items()]
        
    if (r == 0) and (g == 0) and (b == 0):
        r = 1
        g = 1
        b = 1
        
    minDist = 500
    index = 1
    lab = _rgb_to_cielab(float(r), float(g), float(b))
    for i in range(0, len(dxfColors)):
        labPal = _rgb_to_cielab(float(dxfColors[i][0]), float(dxfColors[i][1]), float(dxfColors[i][2]))
        distL = lab[0] - labPal[0]
        dista = lab[1] - labPal[1]
        distb = lab[2] - labPal[2]                
        dist = math.sqrt(distL*distL + dista*dista + distb*distb)
        if dist < minDist:
            minDist = dist
            index = i
    return index

class DXFWriter(StrokeShader):
	def __init__(self, f, w, h):
		StrokeShader.__init__(self)
		self.width, self.height = w, h
		# create the file 
		self.dxf = dxfLibrary130.Drawing(extmax=(self.width,self.height,0))
		self.file = f
	def close(self):
		try:
			self.dxf.saveas(self.file)
			print("dxf saved to file %r" % self.file)
		except:
			print("Cannot save to file %r" % self.file)

	def isPointOutsideViewport(self, point): 
		x, y = point
		if x < 0:
			return CLIP_LEFT
		if y < 0:
			return CLIP_BOTTOM
		if x > self.width:
			return CLIP_RIGHT
		if y > self.height:
			return CLIP_TOP
		return None
			
	def shade(self, stroke):

		points = []
		for v in stroke:
			x, y = v.point
			points.append((x, y))
		
		points_clipped = []			
		if len(stroke) > 0:
			last_clip = self.isPointOutsideViewport(points[0])
			if last_clip == None:
				points_clipped.append(points[0])

		for i in range(1, len(points)):
			clip = self.isPointOutsideViewport(points[i])
			if (clip == None) and (last_clip == None):
				points_clipped.append(points[i])
			else:
				clip_tmp = last_clip
				if (last_clip == None):
					clip_tmp = clip
				if clip_tmp == CLIP_LEFT:
					clip_p1 = (0,0)
					clip_p2 = (0,self.height)
				elif clip_tmp == CLIP_RIGHT:
					clip_p1 = (self.width,0)
					clip_p2 = (self.width,self.height)
				elif clip_tmp == CLIP_TOP:
					clip_p1 = (0,self.height)
					clip_p2 = (self.width,self.height)
				elif clip_tmp == CLIP_BOTTOM:
					clip_p1 = (0,0)
					clip_p2 = (self.width,0)
				intersect = mathutils.geometry.intersect_line_line_2d(mathutils.Vector(clip_p1), mathutils.Vector(clip_p2), mathutils.Vector(points[i-1]), mathutils.Vector(points[i]))
				if intersect:
					points_clipped.append((intersect.x, intersect.y))
					if last_clip != None:
						points_clipped.append(points[i])					
				last_clip = clip
		
		r, g, b = v.attribute.color
		if len(points_clipped) > 0:
			self.dxf.append(dxfLibrary130.PolyLine(points_clipped, color = col2DXF(r, g, b)))


scene = freestyle.getCurrentScene()
f = os.path.join(bpy.path.abspath(scene.render.filepath), "Frame%06d.dxf" % scene.frame_current)
w = scene.render.resolution_x * scene.render.resolution_percentage / 100
h = scene.render.resolution_y * scene.render.resolution_percentage / 100

upred = QuantitativeInvisibilityUP1D(0)
Operators.select(upred)
#Operators.bidirectionalChain(ChainSilhouetteIterator(), NotUP1D(upred))
writer = DXFWriter(f, w, h)
shaders_list = [
	ConstantThicknessShader(2),
	#pyDepthDiscontinuityThicknessShader(1, 4),
	#ConstantColorShader(0, 0, 0),
	pyMaterialColorShader(1),
	writer,
]
Operators.create(TrueUP1D(), shaders_list)
writer.close()

